#!/usr/bin/env python
# encoding: utf-8
#
# Copyright (C) 2020 Henrik tom Wörden
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


import logging

import numpy as np
import pandas as pd
from caosadvancedtools.suppressKnown import SuppressKnown
from xlrd import XLRDError

from .datainconsistency import DataInconsistencyError

logger = logging.getLogger("caosadvancedtools")


def name_converter(name):
    name = str(name)

    if len(name.split(",")) != 2:
        raise ValueError("Name field should be 'LastName, FirstName'."
                         "The supplied value was '{}'.".format(name))

    return name


def yes_no_converter(val):
    if str(val).lower() == "yes":
        return True
    elif str(val).lower() == "no":
        return False
    else:
        raise ValueError(
            "Field should be 'Yes' or 'No', but is '{}'.".format(val))


class XLS_Importer(object):
    def __init__(self, converters, obligatory_columns=[], unique_columns=[]):
        """

        converters: dict with column names as keys and converter functions as
                    values
                    This dict also defines what columns are required to exist
                    throught the existing keys. The converter functions are
                    applied to the cell values. They should also check for
                    ValueErrors, such that a separate value check is not
                    necessary.
        obligatory_columns: list of column names, optional
                            each listed column must not have missing values
        """
        self.sup = SuppressKnown()
        self.required_columns = list(converters.keys())
        self.obligatory_columns = obligatory_columns
        self.unique_columns = unique_columns
        self.converters = converters

    def read_xls(self, filename):
        try:
            xls_file = pd.io.excel.ExcelFile(filename)
        except XLRDError as e:
            logger.warning(
                "Cannot read \n{}.\nError:{}".format(filename,
                                                     str(e)),
                extra={'identifier': str(filename),
                       'category': "inconsistency"})
            raise DataInconsistencyError(*e.args)

        if len(xls_file.sheet_names) > 1:
            # Multiple sheets is the default now. Only show in debug
            logger.debug("Excel file {} contains multiple sheets. "
                         "All but the first are being ignored.".format(
                             filename
                         ))

        try:
            df = xls_file.parse(converters=self.converters)
        except Exception as e:
            logger.warning(
                "Cannot parse {}.".format(filename),
                extra={'identifier': str(filename),
                       'category': "inconsistency"})
            raise DataInconsistencyError(*e.args)

        self.check_columns(df, filename=filename)
        df = self.check_missing(df, filename=filename)

        if len(self.unique_columns) > 0:
            df = self.check_unique(df, filename=filename)

        return df

    def check_columns(self, df, filename=None):
        for col in self.required_columns:
            if col not in df.columns:
                errmsg = "Column '{}' missing in ".format(col)
                errmsg += ("\n{}.\n".format(filename) if filename
                           else "the file.")
                errmsg += "Stopping to treat this file..."
                logger.warning(
                    errmsg,
                    extra={'identifier': str(filename),
                           'category': "inconsistency"})
                raise DataInconsistencyError(errmsg)

    def check_unique(self, df, filename=None):
        df = df.copy()
        uniques = []

        for index, row in df.iterrows():
            element = tuple(row.loc[key] for key in self.unique_columns)

            if element in uniques:
                errmsg = (
                    "The {}. row contains the values '{}'.\nThis value "
                    "combination should be unique, but was used in a previous "
                    "row in\n").format(index+1, element)
                errmsg += "{}.".format(filename) if filename else "the file."
                errmsg += "\nThis row will be ignored!"

                logger.warning(errmsg, extra={'identifier': filename,
                                              'category': "inconsistency"})
                df = df.drop(index)
            else:
                uniques.append(element)

        return df

    def check_missing(self, df, filename=None):
        df = df.copy()

        for index, row in df.iterrows():
            # if none of the relevant information is given, skip

            if np.array([pd.isnull(row.loc[key]) for key in
                         self.obligatory_columns]).all():

                df = df.drop(index)

                continue

            # if any of the relevant information is missing, report it

            for key in self.obligatory_columns:

                if pd.isnull(row.loc[key]):
                    errmsg = (
                        "Required information is missing ({}) in {}. row"
                        " (without header) of "
                        "file:\n{}".format(key, index+1, filename))

                    logger.warning(errmsg, extra={'identifier': filename,
                                                  'category': "inconsistency"})
                    df = df.drop(index)

                    continue

        return df
